// graphics libraries
#include "config.h"
#include "tft_master.h"
#include "tft_gfx.h"
#include <stdlib.h>
#include "math.h"
// threading library
#include <plib.h>
// config.h sets 40 MHz
#define	SYS_FREQ 40000000
#include "pt_cornell_TFT.h"
int PushState = 0;// initial status for debounce
int value; // for test mode
char buffer[60];// string buffer
static struct pt  pt_key, pt_dds;// thread control structs
int i , freq1,freq2, flag1 ; // for sound generation
int row =0, col = 0;  //organize the displaying numbers
int generate_period = 200;  // generate period
int sys_time_seconds ;// system 1 second interval tick
int temp[12] , counter=0,n=0, countert,it; // for play back
volatile signed int sine_table[256]; // dds sine table
volatile unsigned int DDS_accumulator1 = 0,DDS_accumulator2 = 0, DDS_increment1, DDS_increment2 ;// DDS variables
unsigned int sine_index1,sine_index2 ;// sine lookup table
#define DAC_config_chan_A 0b0011000000000000 //writing commend of DAC, enable channel 1 
int rampcounter = 0;// for ramp function
int ramp = 4; //for ramp function 
volatile unsigned int DAC_data ;// output value
volatile SpiChannel spiChn = SPI_CHANNEL2 ;	// the SPI channel to use
volatile int spiClkDiv = 2 ; // 20 MHz max speed for this DAC

// === Interrupt Service Routine =================================================
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
{
    //==========================================================================
    //     Signal output
    //==========================================================================
    mT2ClearIntFlag();
    
    DDS_accumulator1 += DDS_increment1 ;
    sine_index1 = DDS_accumulator1 >> 24 ;//calculate index for dds sine table 1
    DDS_accumulator2 += DDS_increment2 ;
    sine_index2 = DDS_accumulator2 >> 24 ;//calculate index for dds sine table 2
    DAC_data =(sine_table[sine_index1]+sine_table[sine_index2])>>ramp;// calculate ultimate output data of DAC
    
    // CS low to start transaction
    mPORTBClearBits(BIT_4); // start transaction
    
    // test for ready
    while (TxBufFullSPI2());
    
    // write to spi2
    WriteSPI2(DAC_config_chan_A | (DAC_data + 2048));// DAC output commend 
    
    //==========================================================================
    //      Ramp Generation
    //==========================================================================
    if (rampcounter < 1100 && rampcounter % 250 == 0 && ramp >0 ) ramp--;
    else if( rampcounter > 18900  && rampcounter % 250 == 0  && ramp <4 && value == 0)ramp++; 
    rampcounter++;
    
    // test for done
    while (SPI2STATbits.SPIBUSY); // wait for end of transaction
    
     // CS high
    mPORTBSetBits(BIT_4); // end transaction
}


// === Keypad Thread =============================================
static PT_THREAD (protothread_key(struct pt *pt))
{
    PT_BEGIN(pt);
    static int keypad, pattern ;
    // order is 0 thru 9 then * ==10 and # ==11
    // no press = -1
    // table is decoded to natural digit order (except for * and #)
    // 0x80 for col 1 ; 0x100 for col 2 ; 0x200 for col 3
    // 0x01 for row 1 ; 0x02 for row 2; etc
    static int keytable[12]={0x108, 0x81, 0x101, 0x201, 0x82, 0x102, 0x202, 0x84, 0x104, 0x204, 0x88, 0x208};
    // init the keypad pins A0-A3 and B7-B9
    // PortA ports as digital outputs
    mPORTASetPinsDigitalOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);    //Set port as output
    
    // PortB as inputs
    mPORTBSetPinsDigitalIn(BIT_7 | BIT_8 | BIT_9);    //Set port as input

    while(1) 
    {
        //==========================================================================
        //     Key Detection
        //==========================================================================
        
        // read each row sequentially
        mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
        pattern = 1; 
        mPORTASetBits(pattern);
        
        // yield time
        PT_YIELD_TIME_msec(30);
   
        for (i=0; i<4; i++) {
            keypad  = mPORTBReadBits(BIT_7 | BIT_8 | BIT_9);
            if(keypad!=0) {keypad |= pattern ; break;}
            mPORTAClearBits(pattern);
            pattern <<= 1;
            mPORTASetBits(pattern);
        }

        // search for keycode
        if (keypad > 0){ // then button is pushed
            for (i=0; i<12; i++){
                if (keytable[i]==keypad) break;
            }
        }
        else i = -1; // no button pushed

        //==========================================================================
        //     Debounce & Display
        //==========================================================================
        switch (PushState) {
            case 0://NoPush
                if (i!=-1) PushState=1;//MaybePush;
                else PushState=0;//NoPush;
                 break;

            case 1://MaybePush:
                if (i!=-1) 
                {
                    PushState=2;//Pushed;   
                    tft_setTextColor(ILI9340_YELLOW); tft_setTextSize(7);
                    tft_setCursor(row, col);
                    
                    //when press 0--9 :recored
                    if(i<10 && counter<12)//set the max display number to 12
                    {   
                        sprintf(buffer,"%d", i); 
                        tft_writeString(buffer); //display the number being detected
                        row = row+40;
                        
                        if(row%240 == 0)
                        {
                            row =0;
                            col +=60;
                        }// organizing the screen
                        
                        tft_fillRoundRect(0,200,240, 60, 1, ILI9340_BLACK);
                        temp[counter] = i;
                        counter ++;// store the number for playing back    
                    }
                    //set two variables for play back without changing counter and i
                    countert = counter;
                    it = i;
                    
                    //when press * : clear
                    if(i == 10)
                    {
                        tft_fillRoundRect(0,0,240, 300, 0, ILI9340_BLACK);
                        countert = counter = 0;
                        it = i =  0;
                        row = col = 0;
                        //reset column and row 
                        tft_setTextColor(ILI9340_GREEN); tft_setTextSize(7);
                        tft_setCursor(0, 200);
                        sprintf(buffer,"%s", "Clear");
                        tft_writeString(buffer);
                    }
                    flag1 = 1; //enable dds thread and output sound
             
                }
                else PushState=0; //NoPush;
                break;

            case 2://Pushed:  
                if (i!=-1) PushState=2;//Pushed; 
                else PushState=3;//MaybeNoPush;    
                break;

            case 3://MaybeNoPush:
                if (i!=-1) PushState=2;//Pushed; 
                 else PushState=0;//NoPush;    
                break;
        } 
      // NEVER exit while
    } // END WHILE(1)
    PT_END(pt);
} // keypad thread

//dds thread for sound generation and play back
static PT_THREAD (protothread_dds(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1)
    {   
        //==========================================================================
        //     dds sinewave generation
        //==========================================================================
        PT_WAIT_UNTIL(pt, flag1 == 1 );
        if(value == 0)
        {
            //frequencies for normal mode
            switch (i){
                case 1:
                   freq1 = 697;
                   freq2 = 1209;
                   break;
                case 2:
                   freq1 = 697;
                   freq2 = 1336;
                   break;
                case 3:
                   freq1 = 697;
                   freq2 = 1477;
                   break;
                case 4:
                   freq1 = 770;
                   freq2 = 1209;
                   break;
                case 5:
                   freq1 = 770;
                   freq2 = 1336;
                   break;
                case 6:
                   freq1 = 770;
                   freq2 = 1477;
                   break;
                case 7:
                   freq1 = 862;
                   freq2 = 1209;
                   break;
                case 8:
                   freq1 = 862;
                   freq2 = 1336;  
                   break;
                case 9:
                   freq1 = 862;
                   freq2 = 1477;
                   break;
                case 10:
                   freq1 = 941;
                   freq2 = 1209;  
                   break;
                case 0:
                   freq1 = 941;
                  freq2 = 1336;
                   break;
                case 11:
                   freq1 = 941;
                   freq2 = 1477;
                   break;
               default: freq1 = 697;
                        freq2 = 1209;                  
            }
        }
        else
        {
            //frequencies for test mode 
            switch (i){
                case 1:
                   freq1 = 697;
                   freq2 = 0;
                   break;
                case 2:
                   freq1 = 770;
                   freq2 = 0;
                   break;
                case 3:
                   freq1 = 862;
                   freq2 = 0;
                   break;
                case 4:
                   freq1 = 941;
                   freq2 = 0;
                   break;
                case 5:
                   freq1 = 1209;
                   freq2 = 0;
                   break;
                case 6:
                   freq1 = 1336;
                   freq2 = 0;
                   break;
                case 7:
                   freq1 = 1477;
                   freq2 = 0;
                   break;
                default: freq1 = 697;
                         freq2 = 0;       
            }
        }
        
        //calculate two increments for dds to generate sinewave 200k sample rate
        DDS_increment1 = (unsigned int)(freq1*pow(2,32)/(40000000/generate_period));
        DDS_increment2 = (unsigned int)(freq2*pow(2,32)/(40000000/generate_period));
           
        // set up the timer interrupt with a priority of 2 
        OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, generate_period);
        ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2); 
        mT2ClearIntFlag(); //  clear the interrupt flag
        PT_YIELD_TIME_msec(100);  
        
        //extend the latency for test mode
        while (value != 0 )
        {
            value = mPORTAReadBits(BIT_4);
        }            
        CloseTimer2();
        
        //initial ramp and ramp counter for next ramp
        ramp = 4 ;
        rampcounter = 0 ;           
        
        
        //==========================================================================
        //     play back
        //==========================================================================
        if(it!=11)
            {
                flag1 = 0 ;// turn down dds thread
            }
        
        //recall all stored numbers and play them for play back function 
        else if(it == 11 && countert != 0)
            {
                PT_YIELD_TIME_msec(65);
                i = temp[n];
                tft_setTextColor(ILI9340_RED); tft_setTextSize(7);
                tft_setCursor(0, 200);
                sprintf(buffer,"%s", "Recall");
                tft_writeString(buffer);
                n++;
                countert--;
            }
        
        //init all variables when play back finished
        else if( countert == 0) 
        {
            flag1 = 0;
            n = 0;
        }
    }
    PT_END(pt);
}

// === Main  ======================================================
void main(void) {
   // SYSTEMConfigPerformance(PBCLK);
    SYSTEMConfig(SYS_FREQ, SYS_CFG_WAIT_STATES | SYS_CFG_PCACHE);
    
    //set up spi2
    SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV , spiClkDiv);
    
    // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
    PPSOutput(2, RPB5, SDO2);
    
    // init the display 
    ANSELA = 0; ANSELB = 0; CM1CON = 0; CM2CON = 0;
    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9340_BLACK);
    tft_setRotation(0); // Use tft_setRotation(1) for 320x240
    //init protothread
    PT_setup();

    // build sine table
    int i;
        for (i = 0; i < 256; i++)
        {
            sine_table[i] = (signed int)(1024*sin((float)i*6.283/(float)256));  //8 bit amp, 64 samples
        }

    // setup system wide interrupts  ///
    INTEnableSystemMultiVectoredInt();
    
     // control CS for DAC
    mPORTBSetPinsDigitalOut(BIT_4);
    mPORTBSetBits(BIT_4);  
    
    //value for test mode
    mPORTASetPinsDigitalIn(BIT_4);
    CNPUACLR= BIT_4; 
    CNPDASET= BIT_4;
    
    //init the threads
    PT_INIT(&pt_key);
    PT_INIT(&pt_dds);

    // round-robin scheduler for threads
    while (1){
        PT_SCHEDULE(protothread_key(&pt_key));
        PT_SCHEDULE(protothread_dds(&pt_dds));
        value = mPORTAReadBits(BIT_4);
    }
} // main

// === end  ======================================================
